# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.generators
~~~~~~~~~~~~~~~~~~~

This module contains core iterators.

:copyright: (c) 2015 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from .error import ResponseError
from .helpers import get_target_url
from .parsers import (parse_list_objects, parse_list_multipart_uploads,
                      parse_list_parts)

from .signer import sign_v4


class ListObjectsIterator(object):
    """
    Implements list objects iterator for list objects parser.

    :param client: Takes instance of :meth:`urllib3.PoolManager`
    :param url: Target endpoint url where request is served to.
    :param bucket_name: Bucket name resource where request will be served from.
    :param prefix: Prefix name resource for filtering objects.
    :param recursive: Default is non recursive, set True lists all objects
       iteratively.
    :param access_key: Optional if provided requests will be authenticated.
    :param secret_key: Optional if provided requests will be authenticated.
    :param region: Optional if provided requests will be served to this region.
    """
    def __init__(self, client, url, bucket_name, prefix, recursive,
                 access_key=None, secret_key=None, region='us-east-1'):
        self._http = client
        self._endpoint_url = url
        self._bucket_name = bucket_name
        self._prefix = prefix
        self._recursive = recursive
        self._results = []
        self._complete = False
        self._access_key = access_key
        self._secret_key = secret_key
        self._region = region
        self._is_truncated = True
        self._marker = None

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        # if complete, end iteration
        if self._complete:
            raise StopIteration
        # if not truncated and we've emitted everything, end iteration
        if len(self._results) == 0 and self._is_truncated is False:
            self._complete = True
            raise StopIteration
        # perform another fetch
        if len(self._results) == 0:
            self._results, self._is_truncated, self._marker = self._fetch()
        # if fetch results in no elements, end iteration
        if len(self._results) == 0:
            self._complete = True
            raise StopIteration
        # return result
        return self._results.pop(0)

    def _fetch(self):
        query = {}
        query['max-keys'] = 1000
        if self._prefix is not None:
            query['prefix'] = self._prefix
        if not self._recursive:
            query['delimiter'] = '/'
        if self._marker is not None:
            query['marker'] = self._marker

        url = get_target_url(self._endpoint_url,
                             bucket_name=self._bucket_name, query=query)

        method = 'GET'
        headers = {}

        headers = sign_v4(method=method, url=url,
                          region=self._region,
                          headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            response_error = ResponseError(response)
            raise response_error.get(self._bucket_name)

        return parse_list_objects(response.data, bucket_name=self._bucket_name)


class ListIncompleteUploadsIterator(object):
    """
    Implements list incomplete uploads iterator for list multipart uploads
    parser.

    :param client: Takes instance of :meth:`urllib3.PoolManager`
    :param url: Target endpoint url where request is served to.
    :param bucket_name: Bucket name resource where request will be served from.
    :param prefix: Prefix name resource for filtering objects.
    :param delimiter: Default is non recursive, set to *None* to be recursive.
    :param access_key: Optional if provided requests will be authenticated.
    :param secret_key: Optional if provided requests will be authenticated.
    :param region: Optional if provided requests will be served to this region.
    """
    def __init__(self, client, url, bucket_name, prefix=None, delimiter='/',
                 access_key=None, secret_key=None, region='us-east-1'):
        # from user
        self._http = client
        self._endpoint_url = url
        self._bucket_name = bucket_name
        self._prefix = prefix
        self._delimiter = delimiter
        self._access_key = access_key
        self._secret_key = secret_key
        self._region = region

        # internal variables
        self._results = []
        self._complete = False
        self._is_truncated = True
        self._key_marker = None
        self._upload_id_marker = None

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        # if complete, end iteration
        if self._complete:
            raise StopIteration
        # if not truncated and we've emitted everything, end iteration
        if len(self._results) == 0 and self._is_truncated is False:
            self._complete = True
            raise StopIteration
        # perform another fetch
        if len(self._results) == 0:
            self._results, self._is_truncated, self._key_marker, \
                self._upload_id_marker = self._fetch()
        # if fetch results in no elements, end iteration
        if len(self._results) == 0:
            self._complete = True
            raise StopIteration
        # return result
        return self._results.pop(0)

    def _fetch(self):
        query = {
            'uploads': None
        }
        query['max-uploads'] = 1000
        if self._prefix is not None:
            query['prefix'] = self._prefix
        if self._key_marker is not None:
            query['key-marker'] = self._key_marker
        if self._upload_id_marker is not None:
            query['upload-id-marker'] = self._upload_id_marker
        if self._delimiter is not None:
            query['delimiter'] = self._delimiter

        url = get_target_url(self._endpoint_url,
                             bucket_name=self._bucket_name, query=query)

        method = 'GET'
        headers = {}

        headers = sign_v4(method=method, url=url,
                          region=self._region,
                          headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            response_error = ResponseError(response)
            raise response_error.get(self._bucket_name)

        return parse_list_multipart_uploads(response.data,
                                            bucket_name=self._bucket_name)


class ListUploadPartsIterator(object):
    """
    Implements list upload parts iterator for list parts parser.

    :param client: Takes instance of :meth:`urllib3.PoolManager`
    :param url: Target endpoint url where request is served to.
    :param bucket_name: Bucket name resource where request will be served from.
    :param object_name: Object name resource where request will be served from.
    :param upload_id: Upload ID of active multipart to be served.
    :param access_key: Optional if provided requests will be authenticated.
    :param secret_key: Optional if provided requests will be authenticated.
    :param region: Optional if provided requests will be served to this region.
    """
    def __init__(self, client, url, bucket_name, object_name, upload_id,
                 access_key=None, secret_key=None, region='us-east-1'):
        # from user
        self._http = client
        self._endpoint_url = url
        self._bucket_name = bucket_name
        self._object_name = object_name
        self._upload_id = upload_id
        self._access_key = access_key
        self._secret_key = secret_key
        self._region = region

        # internal variables
        self._results = []
        self._complete = False
        self._is_truncated = True
        self._part_marker = None

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        # if complete, end iteration
        if self._complete:
            raise StopIteration
        # if not truncated and we've emitted everything, end iteration
        if len(self._results) == 0 and self._is_truncated is False:
            self._complete = True
            raise StopIteration
        # perform another fetch
        if len(self._results) == 0:
            (self._results,
             self._is_truncated,
             self._part_marker) = self._fetch()
        # if fetch results in no elements, end iteration
        if len(self._results) == 0:
            self._complete = True
            raise StopIteration
        # return result
        potential_result = self._results.pop(0)
        if self._object_name is None:
            return potential_result
        if potential_result.object_name == self._object_name:
            return potential_result
        self._complete = True
        raise StopIteration

    def _fetch(self):
        query = {
            'uploadId': self._upload_id
        }
        query['max-parts'] = 1000
        if self._part_marker is not None:
            query['part-number-marker'] = self._part_marker

        url = get_target_url(self._endpoint_url,
                             bucket_name=self._bucket_name,
                             object_name=self._object_name,
                             query=query)

        method = 'GET'
        headers = {}

        headers = sign_v4(method=method, url=url,
                          region=self._region,
                          headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            response_error = ResponseError(response)
            raise response_error.get(self._bucket_name, self._object_name)

        parts = parse_list_parts(response.data,
                                 bucket_name=self._bucket_name,
                                 object_name=self._object_name,
                                 upload_id=self._upload_id)

        return parts



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 Compatible Cloud Storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
minio.io
~~~~~~~~~~~~~~~

This module contains HTTPReadSeeker implementation which powers
resumable downloads.

:copyright: (c) 2015 by Minio, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from __future__ import absolute_import
import io

from .error import ResponseError


class HTTPReadSeeker(io.IOBase):
    """
    HTTP Read Seeker implements seekable stream.

    This class is also compatible with the Python standard library's :mod:`io`
    module, and can hence be treated as a readable object in the context of
    that framework.

    :param api :class:`Minio <Minio>`
    :param bucket_name: Bucket name of which the object is part of.
    :param object_name: Object name for which :class:`HTTPReadSeeker`
       is created.
    """
    def __init__(self, api, bucket_name, object_name):
        self._api = api
        self._bucket_name = bucket_name
        self._object_name = object_name
        self._is_read = False
        self._stat = None
        self._offset = 0
        self._total_read = 0
        self._reader = None

    def seek(self, offset, whence=0):
        """
        Change the stream position to the given byte *offset*.  *offset* is
        interpreted relative to the position indicated by *whence*. The default
        value for *whence* is :data:`SEEK_SET`.  Values for *whence* are:

        * :data:`SEEK_SET` or ``0`` -- start of the stream (the default);
        *offset* should be zero or positive

        NOT SUPPORTED YET
        ~~~~~~~~~~~~~
        * :data:`SEEK_CUR` or ``1`` -- current stream position; *offset* may
        be negative
        * :data:`SEEK_END` or ``2`` -- end of the stream; *offset* is usually
        negative

        :return: Return the new absolute position.
        """
        # TODO: whence value of '1' and '2' are not implemented yet.
        if offset < 0 and whence == 0:
            raise ValueError('Invalid offset size cannot be negative '
                             'for SEEK_SET')
        self._offset = offset
        return self._offset

    def seekable(self):
        """
        Return ``True`` if the stream supports random access.
        :return: True always.
        """
        # This method is required for `io` module compatibility.
        return True

    def readable(self):
        """
        Return ``True`` if the stream supports read access.
        :return: True always.
        """
        # This method is required for `io` module compatibility.
        return True

    def readinto(self, buf):
        """
        Read up to len(buf) bytes into bytearray *buf* and return the number
        of bytes read.

        Like :meth:`read`, multiple reads may be issued to the underlying
        raw stream, unless the latter is 'interactive'.

        :param buf: Bytearray to read into.
        :return: Length of the read bytes.
        """
        # This method is required for `io` module compatibility.
        temp = self.read(len(buf))
        temp_length = len(temp)
        if temp_length == 0:
            return 0
        buf[:temp_length] = temp
        self._total_read += temp_length
        return temp_length

    def stream(self, amt=2**20):
        """
        A generator wrapper for the read() method. A call will block until
        ``amt`` bytes have been read from the connection or until the
        connection is closed.
            Raise :exc:`ResponseError` on failure.
        :param amt:
            How much of the content to read. The generator will return up to
            much data per iteration, but may return less.
        """
        if self._is_read is False:
            response = self._api.get_partial_object(self._bucket_name,
                                                    self._object_name,
                                                    self._offset, 0)
            if response.status != 206 and response.status != 200:
                response_error = ResponseError(response)
                raise response_error.get(self._bucket_name, self._object_name)

            self._reader = response
            self._is_read = True

        self._total_read += amt
        return self._reader.stream(amt=amt)

    def read(self, amt=None):
        """
        Similar to :meth:`urllib3.HTTPResponse.read`, but with amt option.
            Raise :exc:`ResponseError` on failure.
        :param amt:
            How much of the content to read.
        """
        data = None
        if self._is_read is False:
            # If reading is not started yet, get a new response reader
            # for a specified offset.
            response = self._api.get_partial_object(self._bucket_name,
                                                    self._object_name,
                                                    self._offset, 0)

            if response.status != 206 and response.status != 200:
                response_error = ResponseError(response)
                raise response_error.get(self._bucket_name, self._object_name)

            self._reader = response
            self._is_read = True

        if amt is None:
            data = self._reader.read()
        else:
            data = self._reader.read(amt)

        self._total_read = len(data)
        return data

    def getsize(self):
        """
        Return the size of the Seekable stream.  Raise :exc:`ResponseError`
        if the file does not exist or is inaccessible.
        """
        self._stat = self._api.stat_object(self._bucket_name,
                                           self._object_name)
        return self._stat.size
